import contextlib

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen as _urlopen
    def urlopen(*args, **kwargs):
        return contextlib.closing(_urlopen(*args, **kwargs))

try:
    from importlib import reload
except ImportError:
    from __builtin__ import reload



# coding: utf-8

import sys
import weakref
import json
import datetime
from random import randint

from objecttools import CachedProperty

from xxkcd._moves import urlopen, reload
from xxkcd import constants


_404_mock = {
    'month': '4', 'num': 404, 'link': '', 'year': '2008', 'news': '',
    'safe_title': '404 not found', 'transcript': '', 'alt': '', 'img': '',
    'title': '404 - Not Found', 'day': '1'
}


def _decode(s):
    """
    Decodes the incorrect encoding used in the xkcd JSON api.

    All of the fields of xkcd comics that use non-ascii characters incorrectly
    encode binary UTF-8 as escaped UTF-8.

    For example, 'é' -> 0xC3 0xA9 -> '\\u00c3\\u00a9' in the JSON, which
    actually encodes 'Ã©'

    This is sometimes done multiple times, like in xkcd(124).

    :param str s: String to decode
    :return: Decoded string
    :rtype: Text
    """
    for _ in range(10):
        try:
            s = bytearray(map(ord, s)).decode('utf-8')
        except (ValueError, UnicodeDecodeError):
            return s
    return s


class xkcd(object):
    """Interface with the xkcd JSON API"""
    __slots__ = ('comic', '__weakref__', '__dict__')

    _cache = {}
    _keep_alive = {}

    def __new__(cls, comic=None, keep_alive=False):
        """
        Wrapper around the xkcd API for the comic

        :param Optional[int] comic: Number of the comic.
            Non-positive or `None` for the latest comic
        :param bool keep_alive: True to not delete the data requested until
            the `.delete()` method is called
        :return: The new xkcd object
        :rtype: xkcd
        :raises TypeError: Non-int given as comic
        """
        if comic is not None:
            try:
                comic = int(comic)
            except (ValueError, TypeError):
                raise TypeError('comic must be an integer')
            if comic <= 0 or comic > cls.latest():
                comic = None
        cached = cls._cache.get(comic, None)
        if cached is not None:
            cached = cached()
            if cached is not None:
                return cached
        self = super(xkcd, cls).__new__(cls)
        self.comic = comic
        cls._cache[comic] = weakref.ref(self)
        if keep_alive:
            cls._keep_alive[comic] = self
        return self

    @CachedProperty
    def _raw_json(self):
        """Raw JSON with a possibly incorrect transcript and alt text"""
        if self.comic == 404:
            return _404_mock
        if self.comic is None:
            url = constants.xkcd.json.latest
        else:
            url = constants.xkcd.json.for_comic(number=self.comic)
        with urlopen(url) as http:
            return json.load(http)

    _raw_json.can_delete = True

    @CachedProperty
    def json(self):
        """
        JSON with the correct transcript and encoded properly.

        Note, that in Python 2, all of the `str` keys are `unicode`.

        Format::
        {
            'month': (the number of the month it was published. Possibly None.),
            'num': (int of the number of the comic),
            'link': (str of where the comic links to when clicked on, most commonly a larger version),
            'year': (the year it was published. Possibly None.),
            'news': (str of news that would be placed on the news ticker on the comic. Usually empty.),
            'safe_title': (str of url safe title),
            'transcript': (str of the transript. Possibly empty, usually for later comics),
            'alt': (str of the text that appears in the title attribute of the comic. Possibly empty.),
            'img': (str of the direct link to the image. Possibly empty, usually for interactive comics),
            'title': (str of the original title of the comic.),
            'day': (the number of the day it was published. Possibly None.)
        }

        Type stub::
        {
            'month': Optional[int], 'num': int, 'link': Text,
            'year': Optional[int], 'news': Text, 'safe_title': Text,
            'transcript': Text, 'alt': Text, 'img': Text, 'title': Text,
            'day' Optional[int]
        }
        """
        n = self.comic
        if n is None:
            pass
        # These comics messed up the transcripts.
        # They were interactive comics that didn't have a transcript that
        # pushed the transcripts for all other comics 1 or 2 ahead.
        elif n >= 1663:
            n += 3
        elif n >= 1608:
            n += 2
        other_json = self._raw_json.copy()
        if n < self.latest() - 3:
            other_json['transcript'] = _decode(xkcd(n)._raw_json['transcript'])
        else:
            other_json['transcript'] = _decode(other_json['transcript'])
        for text_key in ('alt', 'title', 'safe_title'):
            other_json[text_key] = _decode(other_json[text_key])
        for int_key in ('day', 'month', 'year'):
            if other_json[int_key]:
                other_json[int_key] = int(other_json[int_key])
            else:
                other_json[int_key] = None
        return other_json

    json.can_delete = True

    @property
    def month(self):
        return self.json['month']

    @property
    def link(self):
        return self.json['link']

    @property
    def year(self):
        return self.json['year']

    @property
    def news(self):
        return self.json['news']

    @property
    def safe_title(self):
        return self.json['safe_title']

    @property
    def transcript(self):
        return self.json['transcript']

    @property
    def alt(self):
        return self.json['alt']

    @property
    def img(self):
        return self.json['img']

    def read_image(self):
        """
        Read the raw image from the link and return the data.

        :raises ValueError: Comic doesn't have an image (e.g., it is xkcd(404), or it is interactive.)
        :return: Raw image data
        :rtype: bytes
        """
        if not self.img:
            raise ValueError('Comic {} does not have an image!'.format(self))
        with urlopen(self.img) as f:
            return f.read()

    @property
    def title(self):
        return self.json['title']

    @property
    def day(self):
        return self.json['day']

    @property
    def date(self):
        """
        :return: The day, month and year the comic was published.
        :rtype: datetime.date
        """
        return datetime.date(self.year, self.month, self.day)

    @classmethod
    def latest(cls):
        """
        :return: The number of the latest comic
        :rtype: int
        """
        return xkcd(keep_alive=True)._raw_json['num']

    @classmethod
    def random(cls):
        """
        :return: A random comic
        :rtype: xkcd
        """
        return cls(randint(1, cls.latest()))

    def delete(self):
        """
        Deletes the data associated with this xkcd object so that it is reloaded
        the next time it is requested

        :return: None
        """
        del self._raw_json
        del self.json
        self._keep_alive.pop(self.comic, None)

    @property
    def explainxkcd(self):
        """
        :return: The explain xkcd wiki link for this comic
        :rtype: str
        """
        if self.comic is None:
            return constants.Explainxkcd.latest
        return constants.Explainxkcd.for_comic(number=self.comic)

    @property
    def url(self):
        """
        :return: The url for this comic
        :rtype: str
        """
        if self.comic is None:
            return constants.xkcd.latest
        return constants.xkcd.for_comic(number=self.comic)

    @property
    def mobile_url(self):
        """
        :return: The mobile url for this comic
        :rtype: str
        """
        if self.comic is None:
            return constants.xkcd.mobile.latest
        return constants.xkcd.for_comic(number=self.comic)

    @property
    def num(self):
        return self.json['num']

    def __repr__(self):
        return '{type.__name__}({number})'.format(type=type(self), number=self.comic or '')

    @property
    def next(self):
        """
        Return the next comic. If is the latest comic, return the same comic

        :return: Next comic
        :rtype: xkcd
        """
        if self.comic is None:
            return self
        return type(self)(self.comic + 1)

    @property
    def previous(self):
        """
        Return the previous comic. If is xkcd(1), return self.

        :return: Previous comic
        :rtype: xkcd
        """
        if self.comic is None:
            return type(self)(type(self).latest() - 1)
        if self.comic == 1:
            return self
        return type(self)(self.comic - 1)

    @staticmethod
    def antigravity():
        """???"""
        antigravity = sys.modules.get('antigravity')
        if antigravity is not None:
            return reload(antigravity)
        else:
            import antigravity
        return antigravity

    def __eq__(self, other):
        if isinstance(other, xkcd):
            if other.comic == self.comic:
                return True
            if self.comic is None:
                return self.num == other.comic
            if other.comic is None:
                return other.num == self.comic
            return False
        return NotImplemented

    def __ne__(self, other):
        eq = self.__eq__(other)
        if eq is NotImplemented:
            return NotImplemented
        return not eq



SECURE_PROTOCOL = 'https:'
INSECURE_PROTOCOL = 'http:'

class xkcd(object):
    base = SECURE_PROTOCOL + '//xkcd.com'
    latest = base
    for_comic = (base + '/{number}').format
    class json(object):
        base = SECURE_PROTOCOL + '//xkcd.com'
        _suffix = '/info.0.json'
        latest = base + _suffix
        for_comic = (base + '/{number}' + _suffix).format
    # Or, alternatively, `http://c.xkcd.com/api-0/jsonp/comic/{number}`
    class c(object):
        base = SECURE_PROTOCOL + '//c.xkcd.com'
        class whatif:
            base = SECURE_PROTOCOL + '//c.xkcd.com/whatif'
            news = base + '/news'
    class mobile(object):
        base = SECURE_PROTOCOL + '//m.xkcd.com'
        latest = base
        for_comic = (base + '/{number}').format

class Explainxkcd(object):
    base = INSECURE_PROTOCOL + '//www.explainxkcd.com'
    latest = base
    for_comic = (base + '/{number}').format

class WhatIf(object):
    base = SECURE_PROTOCOL + '//what-if.xkcd.com'